/*
**   IRC services -- mempool.c
**   Copyright (C) 2001-2004 Chris Behrens
**
**   This program is free software; you can redistribute it and/or modify
**   it under the terms of the GNU General Public License as published by
**   the Free Software Foundation; either version 1, or (at your option)
**   any later version.
**
**   This program is distributed in the hope that it will be useful,
**   but WITHOUT ANY WARRANTY; without even the implied warranty of
**   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
**   GNU General Public License for more details.
**
**   You should have received a copy of the GNU General Public License
**   along with this program; if not, write to the Free Software
**   Foundation, Inc., 675 Mass Ave, Cambridge, MA 02139, USA.
*/


#define __MEMPOOL_C__
#include "chanfix.h"

RCSID("$Id: mempool.c,v 1.3 2004/09/05 03:53:17 musirc Exp $");


static int					SIMPNumEntries = 0;
static int					SIMPTotUsed = 0;

int simempool_init(SI_MEMPOOL *mempool, char *name, u_int flags)
{
	if (!mempool)
		return EINVAL;

	memset(mempool, 0, sizeof(SI_MEMPOOL));
	mempool->mp_name = name;
	mempool->mp_flags = flags;

	return 0;
}

void simempool_deinit(SI_MEMPOOL *mempool)
{
	if (!mempool)
		return;
}

void *simempool_alloc(SI_MEMPOOL *mempool, u_int size)
{
	SI_MEMPOOL_ENTRY	*e;

	if (!size)
		return NULL;

	else
	{   
		e = (SI_MEMPOOL_ENTRY *)malloc(sizeof(SI_MEMPOOL_ENTRY)+size);
		if (!e)													   
			return NULL;											  
	}									  

	e->e_datalen = size;
	e->e_mempool = mempool;

	mempool->mp_num_entries++;
	mempool->mp_tot_used += size;
	SIMPNumEntries++;
	SIMPTotUsed += size;

	return e->e_data;
}

void *simempool_calloc(SI_MEMPOOL *mempool, u_int num, u_int size)
{
	char	*ptr = simempool_alloc(mempool, num*size);

	if (!ptr)
		return NULL;
	memset(ptr, 0, num*size);
	return ptr;
}

char *simempool_strdup(SI_MEMPOOL *mempool, char *string)
{
	u_int	len = strlen(string);
	char	*ptr;

	ptr = simempool_alloc(mempool, len+1);
	if (!ptr)
		return NULL;
	return strcpy(ptr, string);
}


void *simempool_realloc(SI_MEMPOOL *mempool, void *addr, u_int size)
{
	if (!addr)
	{
		return simempool_alloc(mempool, size);
	}

	if (!size)
	{
		simempool_free(addr);
		return NULL;
	}

	simempool_free(addr);

	return simempool_alloc(mempool, size);
}

void simempool_free(void *arg)
{
	SI_MEMPOOL_ENTRY *e;

	e = (SI_MEMPOOL_ENTRY *)(((char *)arg) - offsetof(SI_MEMPOOL_ENTRY, e_data));

	e->e_mempool->mp_tot_used -= e->e_datalen;
	--e->e_mempool->mp_num_entries;
	SIMPTotUsed -= e->e_datalen;
	--SIMPNumEntries;

	free(e);
}
