/* MChanfix -- conn.h
 * Copyright (C) 2004 MChanfix Development
 * $Id: conn.h,v 1.3 2004/08/30 00:00:21 musirc Exp $
 */

#ifndef __CONN_H__
#define __CONN_H__

#include <sys/poll.h>
#include <sys/types.h>
#include "chanfix.h"
#include "queue.h"

typedef struct _si_conntype SI_CONNTYPE;
typedef struct _si_conn	SI_CONN;
typedef struct _si_conninfo SI_CONNINFO;

struct _si_conntype
{
  int (*ct_accept_cb)(SI_CONN *);
  int (*ct_connect_cb)(SI_CONN *);
  int (*ct_read_cb)(SI_CONN *);
  int (*ct_write_cb)(SI_CONN *);
};

struct _si_conn
{
  int con_fd;
  u_int con_pfd_pos;
  struct sockaddr_in con_rem_sin;
  time_t con_firsttime;
  time_t con_lasttime;
  char con_inbuffer[SIMAXPACKET*2];
  char *con_addpos;
  char *con_bufpos;
  char *con_passwd;
  SIQUEUE con_outqueue;

#define SI_CON_TYPE_UNKNOWN 0
#define SI_CON_TYPE_LISTENER 1
#define SI_CON_TYPE_CLIENT 2
  u_int con_type;
#define ConIsUnknown(__x) ((__x)->con_type == SI_CON_TYPE_UNKNOWN)
#define ConIsListener(__x) ((__x)->con_type == SI_CON_TYPE_LISTENER)
#define ConIsClient(__x) ((__x)->con_type == SI_CON_TYPE_CLIENT)
#define SI_CON_FLAGS_DEAD 0x001
#define SI_CON_FLAGS_EAT 0x002 /* eat til newline */
#define SI_CON_FLAGS_CONNECTING 0x004 /* connect() in progress */
#define SI_CON_FLAGS_CONNECTED 0x008
#define SI_CON_FLAGS_TS 0x010
#define SI_CON_FLAGS_PINGSENT 0x020
#define ConIsConnecting(__x) ((__x)->con_flags & SI_CON_FLAGS_CONNECTING)
#define ConDoesTS(__x) ((__x)->con_flags & SI_CON_FLAGS_TS)
  u_int con_flags;
#define SI_CAPAB_VALUE_NOQUITSTORM 0x001
  u_int con_capab_flags;
  SI_CONNINFO *con_conninfo;
  void *con_arg;
  SI_CONN *con_prev;
  SI_CONN *con_next;
};

struct _si_conninfo
{
  struct pollfd *ci_pfds;
  u_int ci_pfds_num;
  u_int ci_pfds_size;
  SI_CONN *ci_conns_list;
  SI_CONN **ci_conns_fd_array;
  u_int ci_conns_fd_size;
  u_int ci_timeout;
  time_t ci_next_check;
};

#define siconn_fd(__x) (__x)->con_fd
#define siconn_arg(__x) (__x)->con_arg
#define siconn_passwd(__x) (__x)->con_passwd
#define siconn_name(__x) (ConIsClient(__x) ? \
			  ((SI_CLIENT *)(__x)->con_arg)->c_name : \
			  ConIsListener(__x) ? "*listener*" : \
			  ConIsUnknown(__x) ? "" : NULL)
#define siconn_ipport(__x) \
		      inet_ntoa((__x)->con_rem_sin.sin_addr), \
		      ntohs((__x)->con_rem_sin.sin_port)

int siconn_add(SI_CONNINFO *, int, int, u_int, void *, SI_CONN **);
int siconn_del(SI_CONN *, char *);
int siconn_poll(SI_CONNINFO *, u_int, char *, u_int);
int siconn_printfv(SI_CONN *, char *, va_list);
int siconn_printf(SI_CONN *, char *, ...);
int siconn_flush(SI_CONN *, struct iovec *, u_int);

#endif
