/* MChanfix -- module.h
 * Copyright (C) 2004 MChanfix Development
 * $Id: module.h,v 1.2 2004/08/30 00:00:21 musirc Exp $
 */

#ifndef __MODULE_H__
#define __MODULE_H__

#include "hash.h"
#include "mempool.h"

typedef int (*si_modcmd_func_t)(SI_CLIENT *from, SI_CLIENT *to, char *args);

#define SIMODULE_CUR_VERSION 1
#define SIMODULE_MIN_VERSION 1
#define SIMODULE_COMPAT_VERSION 1
typedef struct si_mod_client SI_MOD_CLIENT;
typedef struct si_mod_command SI_MOD_COMMAND;
typedef struct si_mod_callbacks SI_MOD_CALLBACKS;

struct si_mod_client
{
  char *mcl_name;
  char *mcl_username;
  char *mcl_hostname;
  char *mcl_info;
  u_char mcl_type;
  u_int mcl_umodes;
  u_int mcl_flags;
  SI_MOD_COMMAND *mcl_commands;
};

struct si_mod_command
{
  char *mc_name;
  char *mc_desc;
#define SI_MC_FLAGS_NOHELP 0x001
#define SI_MC_FLAGS_OPERONLY 0x002
#define SI_MC_FLAGS_NOLOG 0x004
  u_int mc_flags;
  si_modcmd_func_t mc_func;
};

struct si_mod_callbacks
{
  void (*cb_client_added)(SI_CLIENT *);
  void (*cb_server_added)(SI_CLIENT *);
  void (*cb_client_removed)(SI_CLIENT *, char *);
  void (*cb_server_removed)(SI_CLIENT *, char *);
  void (*cb_channel_added)(SI_CHANNEL *);
  void (*cb_channel_removed)(SI_CHANNEL *);
  void (*cb_nick_changed)(SI_CLIENT *, char *);
  void (*cb_umode_changed)(SI_CLIENT *, u_int);
  void (*cb_chmode_changed)(SI_CHANNEL *, u_int, char *, u_int);
  void (*cb_topic_changed)(SI_CHANNEL *, char *, char *, time_t);
  void (*cb_channel_client_added)(SI_CHANNEL *, SI_CLIENT *, SI_CLIENTLIST *);
  void (*cb_channel_client_removed)(SI_CHANNEL *, SI_CLIENT *, SI_CLIENTLIST *);
  void (*cb_channel_clmode_changed)(SI_CHANNEL *, SI_CLIENTLIST *, u_int);
  void (*cb_channel_ban_added)(SI_CHANNEL *, SI_CLIENT *, SI_BAN *);
  void (*cb_channel_ban_removed)(SI_CHANNEL *, SI_CLIENT *, SI_BAN *);
  void (*cb_sjoin_done)(SI_CHANNEL *);
};

struct si_module_decl
{
  u_int mod_version;
  u_int mod_compat_version;
  char *mod_descr;
  SI_MEMPOOL *mod_mempool;
  SI_MOD_CALLBACKS *mod_callbacks;
  int (*mod_init)(void *, char *);
  void (*mod_deinit)(void);

};

#define SI_MODULE_DECL(__a, __b, __c, __d, __e) \
struct si_module_decl simodule_decl = {	\
  SIMODULE_CUR_VERSION,	\
  SIMODULE_COMPAT_VERSION, \
  __a, __b, __c, __d, __e \
}

int simodule_init(void);
void simodule_deinit(void);
int simodule_load(char *, char *, const char **);
int simodule_unload(char *, const char **);
int simodule_data_add(void *, void *, void *);
int simodule_data_get(void *, void *, void **);
int simodule_data_del(void *, void *);
int simodule_add_client(void *, SI_MOD_CLIENT *);
void simodule_del_client(void *, SI_MOD_CLIENT *);
void simodule_client_added(SI_CLIENT *);
void simodule_client_removed(SI_CLIENT *, char *);
void simodule_channel_added(SI_CHANNEL *);
void simodule_channel_removed(SI_CHANNEL *);
void simodule_nick_changed(SI_CLIENT *, char *);
void simodule_umode_changed(SI_CLIENT *, u_int);
void simodule_chmode_changed(SI_CHANNEL *, u_int, char *, int);
void simodule_topic_changed(SI_CHANNEL *, char *, char *, time_t);
void simodule_channel_client_added(SI_CHANNEL *, SI_CLIENT *, SI_CLIENTLIST *);
void simodule_channel_client_removed(SI_CHANNEL *, SI_CLIENT *, SI_CLIENTLIST *);
void simodule_channel_clmode_changed(SI_CHANNEL *, SI_CLIENTLIST *, u_int);
void simodule_channel_ban_added(SI_CHANNEL *, SI_CLIENT *, SI_BAN *);
void simodule_channel_ban_removed(SI_CHANNEL *, SI_CLIENT *, SI_BAN *);
void simodule_sjoin_done(SI_CHANNEL *);
int simodule_process_message(SI_CLIENT *, SI_CLIENT *, char *, int);
#endif
