/*
**   line based config file parsing library -- csconfig_test.c
**   Copyright (C) 2001-2004 Chris Behrens
**
**   This program is free software; you can redistribute it and/or modify
**   it under the terms of the GNU General Public License as published by
**   the Free Software Foundation; either version 1, or (at your option)
**   any later version.
**
**   This program is distributed in the hope that it will be useful,
**   but WITHOUT ANY WARRANTY; without even the implied warranty of
**   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
**   GNU General Public License for more details.
**
**   You should have received a copy of the GNU General Public License
**   along with this program; if not, write to the Free Software
**   Foundation, Inc., 675 Mass Ave, Cambridge, MA 02139, USA.
*/

#include <stdio.h>
#include <stdlib.h>
#include <sys/types.h>

#include "csconfig.h"


static int _command_block1(void *csc_cookie, char *args, void **cb_arg);
static int _command_block2(void *csc_cookie, char *args, void **cb_arg);
static int _command_block3(void *csc_cookie, char *args, void **cb_arg);
static int _command_cmd1(void *csc_cookie, char *args, void **cb_arg);
static int _command_cmd2(void *csc_cookie, char *args, void **cb_arg);
static int _command_cmd3(void *csc_cookie, char *args, void **cb_arg);
static int _command_cmd4(void *csc_cookie, char *args, void **cb_arg);

static CSC_COMMAND	Block3Commands[] = {
	{	"cmd4",		_command_cmd4,		0,		NULL				},
	{	NULL,		NULL,				0,		NULL				}
};

static CSC_COMMAND	Block2Commands[] = {
	{	"cmd3",		_command_cmd3,		0,		NULL				},
	{	NULL,		NULL,				0,		NULL				}
};

static CSC_COMMAND	Block1Commands[] = {
	{	"cmd2",		_command_cmd2,		0,		NULL				},
	{	"block3",	_command_block3,	0,		&Block3Commands		},
	{	NULL,		NULL,				0,		NULL				}
};

static CSC_COMMAND	Commands[] = {
	{	"block1",	_command_block1, 	0,		&Block1Commands		},
	{	"cmd1",		_command_cmd1, 		0,		NULL				},
	{	"block2",	_command_block2, 	0,		&Block2Commands		},
	{	NULL,		NULL,				0,		NULL				}
};

static int _command_block1(void *csc_cookie, char *args, void **cb_arg)
{
	printf("bl1 called, cb_arg=%x, lineno %d\n", (u_int)*cb_arg,
				csconfig_lineno(csc_cookie));
	return 0;
}

static int _command_block2(void *csc_cookie, char *args, void **cb_arg)
{
	printf("bl2 called, cb_arg=%x, lineno %d\n", (u_int)*cb_arg,
				csconfig_lineno(csc_cookie));
	return 0;
}

static int _command_block3(void *csc_cookie, char *args, void **cb_arg)
{
	printf("bl3 called, cb_arg=%x, lineno %d\n", (u_int)*cb_arg,
				csconfig_lineno(csc_cookie));
	*cb_arg = (void *)0x002;
	return 0;
}

static int _command_cmd1(void *csc_cookie, char *args, void **cb_arg)
{
	printf("cmd1 called, cb_arg=%x, lineno %d\n", (u_int)*cb_arg,
				csconfig_lineno(csc_cookie));
	return 0;
}

static int _command_cmd2(void *csc_cookie, char *args, void **cb_arg)
{
	printf("cmd2 called, cb_arg=%x, lineno %d\n", (u_int)*cb_arg,
				csconfig_lineno(csc_cookie));
	return 0;
}

static int _command_cmd3(void *csc_cookie, char *args, void **cb_arg)
{
	printf("cmd3 called, cb_arg=%x, lineno %d\n", (u_int)*cb_arg,
				csconfig_lineno(csc_cookie));
	return 0;
}

static int _command_cmd4(void *csc_cookie, char *args, void **cb_arg)
{
	printf("cmd4 called, cb_arg=%x, lineno %d\n", (u_int)*cb_arg,
				csconfig_lineno(csc_cookie));
	return 0;
}


int main(int argc, char **argv)
{
	int		err;

	if (argc < 2)
	{
		fprintf(stderr, "Usage: csconfig_test <configfile>\n");
		exit(-1);
	}

	err = csconfig_read(argv[1], Commands, (void *)0x001, NULL);

	printf("csconfig_read() returned: %d\n", err);

	exit(err);
}
